
/* ---------------------------- float2wav16.c ------------------------------ */

/* This program reads raw float samples from stdin, converts to short int */
/* (w/ dithering), and writes a WAV format soundfile			  */

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <math.h>
#include <float.h>
#include <errno.h>
#include <unistd.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <fcntl.h>

#define	BLKSIZE		4096

char	*prgname, *fname;	/* program name, output file name  */
int	nchnls, srate;		/* number of channels, sample rate */
int	fpos, fsize, fsize2;	/* file position and size	   */
int	outfile;		/* output file			   */
float	in_buf[BLKSIZE];	/* input buffer			   */
signed short int
	out_buf[BLKSIZE];	/* output buffer		   */
int	i, j;

char	wavhdr1[5]="RIFF";	/* WAV header strings */
int	wavsize1 = 36;
char	wavhdr2[5]="WAVE";
char	wavhdr3[5]="fmt ";
int	wavfmtsize = 16;
short int
	wavformat = 0x0001;
short int
	wavchnls = 1;
int	wavsrate = 48000;
int	wavbytes_per_sec = 96000;
short int
	wavbytes_per_sample = 2;
short int
	wavbits = 16;
char	wavhdr4[5]="data";
int	wavsize2 = 0;

/* ------------------------------------------------------------------------- */

signed short int
	float2short (float);	/* convert from float to short int */
void	ReadBlock (float*);	/* read BLKSIZE bytes from input   */
void	WriteBlock (int, signed short int*);	/* write BLKSIZE   */
						/* bytes to output */
void	WriteWAVHeader (void);	/* write WAV header		   */
void	WriteFile (int, void*);	/* write to output file		   */

/* convert the float value x to signed short int (also add dither noise) */

signed short int
	float2short (float x)
{
	long int		n;
	float			y;
	signed short int	z;

	n = rand ();
	n -= (RAND_MAX >> 1);
	y = (float) n / (float) (RAND_MAX);
	y += x;
	y += (y > 0.0 ? 0.5 : -0.5);
	y = (y > 32767.0 ? 32767.0 : y);
	y = (y < -32767.0 ? -32767.0 : y);
	z = (signed short int) y;

	return z;
}

/* read BLKSIZE bytes from input */

void	ReadBlock (float *xblock)
{
	int	i;

	for (i = 0; i < (int) (BLKSIZE / sizeof (float)); i++)
		xblock[i] = 0.0;

	i = read (STDIN_FILENO, xblock, BLKSIZE);
	if (i == -1) {
		fprintf (stderr, "%s: error reading stdin: %s\n",
				prgname, strerror (errno) );
		exit (-1);
	}

	fsize2 += (i / (nchnls * sizeof (float)));
	if (i == 0) fsize = fsize2;
}

/* write BLKSIZE bytes to output */

void	WriteBlock (int pos, signed short int *xblock)
{
	int	j;

	/* pos =  file position in samples		*/
	/* j = number of bytes to write to output file	*/

	j = BLKSIZE;
	if (pos > (int) (fsize - (BLKSIZE / (nchnls * sizeof (short int)))))
		j = (fsize - pos) * (nchnls * sizeof (short int));

	WriteFile (j, xblock);
}

/* write WAV header */

void	WriteWAVHeader (void)
{
	int	i;

	i = lseek (outfile, 0, SEEK_SET);
	if (i == -1) {
		fprintf (stderr, "%s: cannot seek in %s: %s\n",
				prgname, fname, strerror (errno) );
		exit (-1);
	}

	WriteFile (4, wavhdr1);
	WriteFile (4, &wavsize1);
	WriteFile (4, wavhdr2);
	WriteFile (4, wavhdr3);
	WriteFile (4, &wavfmtsize);
	WriteFile (2, &wavformat);
	WriteFile (2, &wavchnls);
	WriteFile (4, &wavsrate);
	WriteFile (4, &wavbytes_per_sec);
	WriteFile (2, &wavbytes_per_sample);
	WriteFile (2, &wavbits);
	WriteFile (4, wavhdr4);
	WriteFile (4, &wavsize2);
}

/* write size bytes to output file */

void	WriteFile (int size, void *xblock)
{
	int	i;

	if (size > 0) {
		i = write (outfile, xblock, size);
		if (i == -1) {
			fprintf (stderr, "%s: error writing to %s: %s\n",
				prgname, fname, strerror (errno) );
			exit (-1);
		}
	}
}

/* ------------------------------------------------------------------------- */

int	main (int argc, char **argv)
{
	prgname = argv[0];

	/* print usage */

	if (argc != 4) {
		fprintf (stderr, "usage: %s <output file name> ", prgname);
		fprintf (stderr, "<number of channels> <sample rate>\n");
		fprintf (stderr, "  This program reads raw float data from ");
		fprintf (stderr, "stdin,\n  converts to short int, and ");
		fprintf (stderr, "writes a WAV format soundfile.\n");
		exit (0);
	}

	/* set sample rate and number of channels */

	nchnls = (int) atoi (argv[2]);
	srate = (int) atoi (argv[3]);

	if ((nchnls != 1) && (nchnls != 2) && (nchnls != 4) && (nchnls != 8)) {
		fprintf (stderr, "%s: ", prgname);
		fprintf (stderr, "number of channels must be 1, 2, 4, or 8\n");
		exit (-1);
	}

	if ((srate < 1) || (srate > 10000000)) {
		fprintf (stderr, "%s: invalid sample rate: %d\n",
				prgname, srate);
		exit (-1);
	}

	/* open output file */

	fname = argv[1];
	outfile = open (fname, O_WRONLY | O_CREAT | O_TRUNC, 0644);
	if (outfile == -1) {
		fprintf (stderr, "%s: ", prgname);
		fprintf (stderr, "error opening %s: %s\n",
				fname, strerror (errno) );
		exit (-1);
	}

	/* fpos = file position, fsize = file size in samples */

	fpos = 0;
	fsize = 0x7FFFFFFF;
	fsize2 = 0;

	srand (16);

	/* write WAV header */

	wavsrate = srate;
	wavbits = 16;
	wavchnls = nchnls;
	wavbytes_per_sample = nchnls * sizeof (signed short int);
	wavbytes_per_sec = wavsrate * wavbytes_per_sample;

	WriteWAVHeader ();

	do {	j = 0;		/* sizeof (float) > sizeof (short int) */
		do {
			ReadBlock (in_buf);
			for (i = 0; i < (int) (BLKSIZE / sizeof (float)); i++)
				out_buf[i+j] = float2short (in_buf[i]);
			j += (BLKSIZE / sizeof (float));
		} while (j < (int) (BLKSIZE / sizeof (signed short int)));

		WriteBlock (fpos, out_buf);
		fpos += (BLKSIZE / (nchnls * sizeof (signed short int)));
	} while (fpos < fsize);

	/* set size in WAV header */

	wavsize2 = fsize * nchnls * sizeof (signed short int);
	wavsize1 = wavsize2 + 36;

	WriteWAVHeader ();

	/* close output file */

	i = close (outfile);
	if (i == -1) {
		fprintf (stderr, "%s: error closing %s: %s\n",
				prgname, fname, strerror (errno) );
		exit (-1);
	}

	return 0;
}


