
/* ------------------------------ maximizer.c ------------------------------ */

/* maximizer / limiter v1.1			*/
/* main sound processing functions		*/
/* written by Istvan Varga 1999, 2000		*/

#include "main.h"
#include "maximizer.h"

/* generate Hanning window */

void	InitWindowTable (int size)
{
	int	i;

	for (i = 0; i < 5120; i++)

		WindowTable[i] = (float) (i < size ?
			pow (sin (PI * (double) i / (double) size), 2.0)
			: 0.0);
}

/* check if there are samples out of range in xblock	*/
/* return value: 0: no samples were out of range	*/
/*		 1: negative peak < -1.0		*/
/*		 2: positive peak > 1.0			*/
/*		 3: both peaks are out of range		*/

int	CheckMaxOut (int size, float *xblock)
{
	int	i, j;

	j = 0;
	for (i = 0; i < size; i++) {
		if (xblock[i] < -1.0) {
			j |= 1;
		} else if (xblock[i] > 1.0) {
			j |= 2;
		}
		if (j == 3) break;
	}

	return j;	/* return 0 if no samples are out of range */
}

/* initialize filter coefficients used by ErrPower */

void	InitFilters (void)
{
	fxb1 = 2.0 - cos(2.0*PI*ffreq1/esr); fxb1 -= sqrt(fxb1*fxb1-1.0);
	fxb2 = 2.0 - cos(2.0*PI*ffreq2/esr); fxb2 -= sqrt(fxb2*fxb2-1.0);
	fxb3 = 2.0 - cos(2.0*PI*ffreq3/esr); fxb3 -= sqrt(fxb3*fxb3-1.0);
	fxb4 = 2.0 - cos(2.0*PI*ffreq4/esr); fxb4 -= sqrt(fxb4*fxb4-1.0);

	fxa1 = 1.0 - fxb1; fxc1 = flvh1; fxd1 = flvl1 - flvh1;
	fxa2 = 1.0 - fxb2; fxc2 = flvh2; fxd2 = flvl2 - flvh2;
	fxa3 = 1.0 - fxb3; fxc3 = flvh3; fxd3 = flvl3 - flvh3;
	fxa4 = 1.0 - fxb4; fxc4 = flvh4; fxd4 = flvl4 - flvh4;
}

/* calculate power of error signal in xblock, when dcoffs offset is added */
/* xblock_ref is the original input signal				  */
/* return value is -1.0 if the error is > ref_err			  */

float	ErrPower (int size, float *xblock, float dcoffs, float *xblock_ref,
		  float ref_err)
{
	int	i;
	float	xerr, a, c, x, y;
	float	f1, f2, f3, f4;

	/* initialize filters */

	f1 = f2 = f3 = f4 = 0.0;

	c = 0.0; y = ref_err * (float) size;

	for (i = 0; i < (size+64); i++) {	/* +64 samples for */
						/* filter response */
		if (i < size) {

			/* add DC offset and clip */

			a = x = xblock[i] + dcoffs;
			x = (x < 0.0 ?
				(x < -1.0 ? (-1.0-x) : 0.0)
			      :	(x > 1.0 ? (1.0-x) : 0.0) );

			/* a = windowed error signal */

			a = WindowTable[i]
				* (xblock_ref[i] - (clip_sens*x + a));

		} else	a = 0.0;

		/* apply filter */

		f1 = fxa1*a + fxb1*f1; a = fxc1*a + fxd1*f1;
		f2 = fxa2*a + fxb2*f2; a = fxc2*a + fxd2*f2;
		f3 = fxa3*a + fxb3*f3; a = fxc3*a + fxd3*f3;
		f4 = fxa4*a + fxb4*f4; a = fxc4*a + fxd4*f4;

		c += a*a;		/* calculate and sum power */

		if (c > y) return -1.0;	/* error > ref_err */
		
	}

	xerr = c / ((float)size);	/* return power of error signal */
	return xerr;
}

/* find the best offset for (size) floats in xblock and return it in xoffs */
/* return value of function is the minimum power of error signal	   */
/* xblock_ref is the original input signal				   */

float	IterateOffset (int size, float *xblock, float *xoffs,
			float *xblock_ref)
{
	float	xerr, xerrmin, offs1, offs2, offsx;
	int	i;

	/* init values for offset and min. error signal */

	offs1 = 0.0;
	xerrmin = ErrPower (size, xblock, offs1, xblock_ref, 1.0e24);
	offsx = offs1;

	do {	offs2 = (*xoffs) * 0.5;		/* offset increment */
		i = 0;				/* iteration number */

		do {	offs2 *= 2.0; i++;	/* increase offset */
						/* exponentially */
			offs1 += offs2;
			offs1 = (offs1 < (-max_offs) ? -max_offs : offs1);
			offs1 = (offs1 > max_offs ? max_offs : offs1);

			xerr = ErrPower (size, xblock, offs1, xblock_ref,
					 xerrmin);
			if (xerr < -0.5) break;	/* error signal */
						/* cannot be reduced */
			if (xerr < xerrmin) {
				xerrmin = xerr;	/* new offset is better than */
				offsx = offs1;	/* previous, so save it */
			}

		/* process until max. offset is reached or error signal */
		/* cannot be improved (best offset is found) */

		} while ((xerr <= xerrmin) &&
			 (fabs (offs1) < (max_offs - offs_0*0.5)));

	} while (i > 1);	/* only 1 iteration -> best offset is found, */
				/* stop processing */

	*xoffs = offsx;		/* return best DC offset and minimum */
	return xerrmin;		/* power of error signal */
}

/* j-stereo version of IterateOffset */

float	IterateOffset_j (int size, float *xblockL, float *xblockR,
			 float *xoffs,
			 float *xblock_ref_L, float *xblock_ref_R)
{
	float	xerr, xerrmin, offs1, offs2, offsx;
	int	i;

	/* init. values for offset and min. error signal */

	offs1 = 0.0;
	xerrmin = ErrPower (size, xblockL, offs1, xblock_ref_L, 1.0e24)
		+ ErrPower (size, xblockR, offs1, xblock_ref_R, 1.0e24);
	offsx = offs1;

	do {	offs2 = (*xoffs) * 0.5;	/* offset increment */
		i = 0;			/* iteration number */

		do {	offs2 *= 2.0; i++;	/* increase offset */
						/* exponentially */
			offs1 += offs2;
			offs1 = (offs1 < (-max_offs) ? -max_offs : offs1);
			offs1 = (offs1 > max_offs ? max_offs : offs1);

			/* sum the power of error signal in L and R channels */

			xerr = ErrPower (size, xblockL, offs1, xblock_ref_L,
					 xerrmin)
			     + ErrPower (size, xblockR, offs1, xblock_ref_R,
			     		 xerrmin);
			if (xerr < -0.5) break;	/* error signal */
						/* cannot be reduced */
			if (xerr < xerrmin) {
				xerrmin = xerr;	/* new offset is better than */
				offsx = offs1;	/* previous, so save it */
			}

		/* process until max. offset is reached or error signal */
		/* cannot be improved (best offset is found) */

		} while ((xerr <= xerrmin) &&
			 (fabs (offs1) < (max_offs - offs_0*0.5)));

	} while (i > 1);	/* only 1 iteration -> best offset is found, */
				/* stop processing */

	*xoffs = offsx;		/* return best DC offset and minimum */
	return xerrmin;		/* power of error signal */
}

/* maximize 1 band in xblockL and xblockR				*/
/* original input signal is stored in xblock_ref_L and xblock_ref_R	*/

int	MaximizeBlock (int size, float *xblockL, float *xblockR,
			float *xblock_ref_L, float *xblock_ref_R)
{
	float	offs1, offs2, errmin1, errmin2, offsx;
	int	i, m, x;

	m = 0;

	/* LEFT channel */

	x = CheckMaxOut (size, xblockL);
	if (x) {
		m = 1; offs1 = offs2 = 0.0; errmin1 = errmin2 = 1.0e24;

		/* find best offset value */

		if (x & 1) {
			offs1 = offs_0;
			errmin1 = IterateOffset (size, xblockL,
						 &offs1, xblock_ref_L);
		}
		if (x & 2) {
			offs2 = -offs_0;
			errmin2 = IterateOffset (size, xblockL,
						 &offs2, xblock_ref_L);
		}

		offsx = (errmin1 < errmin2 ? offs1 : offs2);

		/* and mix to buffer */

		for (i = 0; i < size; i++) xblockL[i] += offsx;
	}

	/* RIGHT channel */

	x = CheckMaxOut (size, xblockR);
	if (x) {
		m = 1; offs1 = offs2 = 0.0; errmin1 = errmin2 = 1.0e24;

		/* find best offset value */

		if (x & 1) {
			offs1 = offs_0;
			errmin1 = IterateOffset (size, xblockR,
						 &offs1, xblock_ref_R);
		}
		if (x & 2) {
			offs2 = -offs_0;
			errmin2 = IterateOffset (size, xblockR,
						 &offs2, xblock_ref_R);
		}

		offsx = (errmin1 < errmin2 ? offs1 : offs2);

		/* and mix to buffer */

		for (i = 0; i < size; i++) xblockR[i] += offsx;
	}

	return m;	/* return 0 if there was nothing to do */
}

/* j-stereo version of MaximizeBlock */

int	MaximizeBlock_j (int size, float *xblockL, float *xblockR,
			 float *xblock_ref_L, float *xblock_ref_R)
{
	float	offs1, offs2, errmin1, errmin2, offsx;
	int	i, x;

	/* there is nothing to do if no samples are out of range */

	x = CheckMaxOut (size, xblockL) | CheckMaxOut (size, xblockR);

	if (x == 0) return 0;

	/* find best offset value (will be the same for both channels */
	/* in case of j-stereo mode) */

	offs1 = offs2 = 0.0; errmin1 = errmin2 = 1.0e24;

	if (x & 1) {
		offs1 = offs_0;
		errmin1 = IterateOffset_j (size, xblockL, xblockR, &offs1,
					   xblock_ref_L, xblock_ref_R);
	}
	if (x & 2) {
		offs2 = -offs_0;
		errmin2 = IterateOffset_j (size, xblockL, xblockR, &offs2,
					   xblock_ref_L, xblock_ref_R);
	}

	offsx = (errmin1 < errmin2 ? offs1 : offs2);

	/* mix to buffers */

	for (i = 0; i < size; i++) {

		xblockL[i] += offsx;
		xblockR[i] += offsx;
	}

	return 1;
}

/* limiter function (independent channels) */

int	LimitBlock (int size, float *xblockL, float *xblockR)
{
	int	i, m;
	float	x;

	m = 0;

	/* LEFT channel */

	if (CheckMaxOut (size, xblockL)) {

		m = 1;

		x = 0.0;		/* calculate RMS */

		for (i = 0; i < size; i++) x += xblockL[i] * xblockL[i];
		x = sqrt (x / (float) size);

		if (x > 1.0) for (i = 0; i < size; i++) xblockL[i] /= x;
	}

	/* RIGHT channel */

	if (CheckMaxOut (size, xblockR)) {

		m = 1;

		x = 0.0;		/* calculate RMS */

		for (i = 0; i < size; i++) x += xblockR[i] * xblockR[i];
		x = sqrt (x / (float) size);

		if (x > 1.0) for (i = 0; i < size; i++) xblockR[i] /= x;
	}

	return m;	/* return 0 if there was nothing to do */
}

/* j-stereo version of LimitBlock */

int	LimitBlock_j (int size, float *xblockL, float *xblockR)
{
	int	i;
	float	x, y;

	if ( ! (    (CheckMaxOut (size, xblockL))
		 || (CheckMaxOut (size, xblockR)) ) ) return 0;

	x = y = 0.0;	/* calculate RMS */

	for (i = 0; i < size; i++) {

		x += xblockL[i] * xblockL[i];
		y += xblockR[i] * xblockR[i];
	}

	/* x = RMS of louder channel */

	x = sqrt ((x > y ? x : y) / (float) size);

	if (x > 1.0)
		for (i = 0; i < size; i++) {

			xblockL[i] /= x;
			xblockR[i] /= x;
		}
	
	return 1;
}

/* apply Hanning window to (size) floats in xblockL and xblockR */

void	WindowBlock (int size, float *xblockL, float *xblockR)
{
	int	i;

	for (i = 0; i < size; i++) {
		xblockL[i] *= WindowTable[i];
		xblockR[i] *= WindowTable[i];
	}
}

/* maximize all bands in xblock (packed (L-R interleaved) stereo data) */

void	MaximizeBlock2 (float *xblock)
{
	int	blksize, i, j, k, m;
	float	tmpblk_ref[(ALLBLKSIZE << 1)];
	float	tmpblk[(ALLBLKSIZE << 1)];
	float	scale1, scale2, scale3;
	float	tmpblockL[5120], tmpblockR[5120];
	float	*tmpblock_ref_L, *tmpblock_ref_R;

	/* k = band number (+2), blksize = frame size */

	k = lband; blksize = MAXBLKSIZE >> k;

	/* save input signal, and uninterleave stereo samples */

	for (i = 0; i < (ALLBLKSIZE << 1); i++)
		tmpblk_ref[((i >> 1) + (ALLBLKSIZE * (i & 1)))] = xblock[i];
	for (i = 0; i < (ALLBLKSIZE << 1); i++)
		xblock[i] = tmpblk_ref[i];

	/* main loop */

	do {	/* pre-scale buffers to set rel. limit */

		scale2 = limit * rel_limit[k]; scale1 = 1.0 / scale2;
		scale3 = 2.0 / ovrlap[k];

		for (i = 0; i < (ALLBLKSIZE << 1); i++) {

			xblock[i] *= scale1;
			tmpblk_ref[i] *= scale1;
			tmpblk[i] = 0.0;	/* clear output buffer */
		}

		/* no samples are out of range, return */

		if (CheckMaxOut ((ALLBLKSIZE << 1), xblock) == 0) {

			for (i = 0; i < (ALLBLKSIZE << 1); i++)
				xblock[i] *= scale2;
			break;
		}

		/* tmpblockL, tmpblockR:	frame data		*/
		/* tmpblk_ref, tmpblock_ref_L,				*/
		/*	       tmpblock_ref_R:	original input signal	*/
		/*				scaled with limit and	*/
		/*				rel_limit		*/

		/* set clip sensitivity and offset parameters */

		clip_sens = clp_sn[k]; max_offs = offs_max[k];
		offs_0 = offs_res[k];

		m = 0;	/* tmp variable to check if there were samples out */
			/* of range in the buffer (processing will be */
			/* stopped if there were none) */

		InitWindowTable (blksize);	/* initialize window table */

		for (i = 0;
		     i <= (ALLBLKSIZE - blksize);
		     i += (blksize / ovrlap[k]) ) {

			/* get frame data */

			for (j = 0; j < blksize; j++) {
				tmpblockL[j] = xblock[i + j];
				tmpblockR[j] = xblock[i + j + ALLBLKSIZE];
			}
			tmpblock_ref_L = tmpblk_ref + i;
			tmpblock_ref_R = tmpblk_ref + i + ALLBLKSIZE;

			/* maximize depending on stereo and limit mode */

			if (l_mode) {			/* limiter mode */

				if (stereo_mode[k])
					m |= LimitBlock_j
						(blksize,
						 tmpblockL, tmpblockR);
				else	m |= LimitBlock
						(blksize,
						 tmpblockL, tmpblockR);

			} else {			/* maximizer mode */
				if (stereo_mode[k])
					m |= MaximizeBlock_j
						(blksize,
						 tmpblockL, tmpblockR,
						 tmpblock_ref_L,
						 tmpblock_ref_R);
				else	m |= MaximizeBlock
						(blksize,
						 tmpblockL, tmpblockR,
						 tmpblock_ref_L,
						 tmpblock_ref_R);
			}

			/* mix windowed signal to output buffer */

			WindowBlock (blksize, tmpblockL, tmpblockR);

			for (j = 0; j < blksize; j++) {
				tmpblk[i + j] += scale3 * tmpblockL[j];
				tmpblk[i + j + ALLBLKSIZE] +=
						 scale3 * tmpblockR[j];
			}
		}

		/* restore original amplitude */

		for (i = 0; i < (ALLBLKSIZE << 1); i++) {

			xblock[i] = scale2 * tmpblk[i];
			tmpblk_ref[i] *= scale2;
		}

		blksize >>= 1; k++;	/* next band */

	/* stop when max. band is reached or no sample was out of range */

	} while ((k < (lband + numbands)) && (m));

	/* restore interleaved format */

	for (i = 0; i < (ALLBLKSIZE << 1); i++)
		tmpblk_ref[i] = xblock[((i >> 1) + (ALLBLKSIZE * (i & 1)))];
	for (i = 0; i < (ALLBLKSIZE << 1); i++)
		xblock[i] = tmpblk_ref[i];

	return;
}


